# Copyright 2024 Niels Provos
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
PlanAI Command Line Interface

This module provides a command-line interface for PlanAI, focusing on automated prompt optimization.

The main functionality includes:
1. Optimizing prompts based on debug logs
2. Configuring LLM providers and models
3. Processing input from Python files, debug logs, and goal prompts
4. Outputting optimized configurations

Usage:
    python -m planai.cli --llm-provider <provider> --llm-model <model> --llm-reason-model <reason_model> optimize-prompt [options]
"""

import argparse
import sys
from typing import List

from planai import __version__, llm_from_config

from .cli_cache import handle_cache_subcommand
from .cli_optimize_prompt import optimize_prompt


def parse_comma_separated_list(arg: str) -> List[str]:
    # Split the argument by commas and strip any extra whitespace
    return [item.strip() for item in arg.split(",")]


def create_parser():
    parser = argparse.ArgumentParser(description="planai command line interface")

    # Global arguments
    parser.add_argument(
        "--llm-provider", type=str, required=False, help="LLM provider name"
    )
    parser.add_argument(
        "--llm-model", type=str, required=False, help="LLM model name for generation"
    )
    parser.add_argument(
        "--llm-reason-model",
        type=str,
        required=False,
        help="LLM model name for reasoning",
    )

    subparsers = parser.add_subparsers(dest="command", help="Subcommands")

    # Subcommand optimize-prompt
    optimize_parser = subparsers.add_parser(
        "optimize-prompt", help="Optimize prompt based on debug logs"
    )
    optimize_parser.add_argument(
        "--python-file", type=str, required=False, help="Path to the Python file"
    )
    optimize_parser.add_argument(
        "--class-name", type=str, required=False, help="Class name in the Python file"
    )
    optimize_parser.add_argument(
        "--debug-log", type=str, required=False, help="Path to the JSON debug log file"
    )
    optimize_parser.add_argument(
        "--goal-prompt", type=str, required=False, help="Goal prompt for optimization"
    )
    optimize_parser.add_argument(
        "--output-config", type=str, help="Output a configuration file"
    )
    optimize_parser.add_argument(
        "--search-path", type=str, help="Optional path to include in module search path"
    )
    optimize_parser.add_argument(
        "--config", type=str, help="Path to a configuration file"
    )
    optimize_parser.add_argument(
        "--num-iterations",
        type=int,
        default=3,
        help="Number of optimization iterations",
    )
    optimize_parser.add_argument(
        "--llm-opt-provider",
        type=str,
        required=False,
        help="LLM provider name to be used for the prompt that is being optimized. This should be the same LLM as being used in production.",
    )
    optimize_parser.add_argument(
        "--llm-opt-model",
        type=str,
        required=False,
        help="LLM model name for generation to be used for the prompt that is being optimized. This should be the same LLM as being used in production.",
    )
    optimize_parser.add_argument(
        "--output-dir",
        type=str,
        default=".",
        help="Directory to write output files",
    )

    # Subcommand cache
    cache_parser = subparsers.add_parser("cache", help="Inspect and manipulate cache")
    cache_parser.add_argument(
        "cache_dir", type=str, help="Directory of the diskcache to operate on"
    )
    cache_parser.add_argument(
        "--output-task-filter",
        type=str,
        help="Filter for output task type",
        default=None,
    )
    cache_parser.add_argument(
        "--delete", type=str, help="Delete a specific cache key", default=None
    )
    cache_parser.add_argument("--clear", action="store_true", help="Clear the cache")
    cache_parser.add_argument(
        "--search-dirs",
        type=parse_comma_separated_list,
        help="Comma-separated list of directories to search for Python modules",
        default=None,
    )

    # Subcommand version
    subparsers.add_parser("version", help="Display PlanAI version")

    return parser


def main(args=None):
    if args is None:
        args = sys.argv[1:]

    parser = create_parser()
    parsed_args = parser.parse_args(args)

    if parsed_args.command == "version":
        print(f"PlanAI version {__version__}")
        return

    if parsed_args.command == "cache":
        handle_cache_subcommand(parsed_args)
        return

    llm_fast = llm_from_config(
        provider=parsed_args.llm_provider, model_name=parsed_args.llm_model
    )
    llm_reason = llm_from_config(
        provider=parsed_args.llm_provider, model_name=parsed_args.llm_reason_model
    )

    if parsed_args.command == "optimize-prompt":
        optimize_prompt(llm_fast, llm_reason, parsed_args)
    else:
        parser.print_help()


if __name__ == "__main__":
    main()



# Copyright 2024 Niels Provos
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import logging
from typing import Any, Dict, List, Optional, Type

from pydantic import Field, PrivateAttr

from .graph import Graph
from .provenance import ProvenanceChain
from .task import Task, TaskWorker

PRIVATE_STATE_KEY = "_graph_task_private_state"


class SubGraphWorkerInternal(TaskWorker):
    graph: Graph = Field(
        ..., description="The graph that will be run as part of this TaskWorker"
    )
    entry_worker: Optional[TaskWorker] = Field(
        None, description="The entry point of the graph"
    )
    exit_worker: Optional[TaskWorker] = Field(
        None, description="The exit point of the graph"
    )
    _state: Dict[str, Any] = PrivateAttr(default_factory=dict)

    def model_post_init(self, _: Any):
        if self.entry_worker is None:
            if len(self.graph.get_entry_workers()) != 1:
                raise ValueError(
                    f"Graph must have exactly one entry worker, got {self.graph.get_entry_workers()}"
                )
            self.entry_worker = self.graph.get_entry_workers()[0]
        if self.exit_worker is None:
            self.exit_worker = self.graph.get_exit_worker()
            if self.exit_worker is None:
                raise ValueError("Graph must have a defined exit worker")
        if len(self.exit_worker.output_types) != 1:
            raise ValueError(
                f"Exit worker must have exactly one output type, got {self.exit_worker.output_types}"
            )

        output_type = self.exit_worker.output_types[0]
        graph_task = self
        self.output_types = [output_type]

        class AdapterSinkWorker(TaskWorker):
            output_types: List[Type[Task]] = [output_type]

            def consume_work(self, task: output_type):
                # we need to move this task from the sub-graph to the main graph
                # first, we need to fix up the provenance so that it shows only the GraphTask as the parent
                old_task = task.get_private_state(PRIVATE_STATE_KEY)
                if old_task is None:
                    raise ValueError(
                        f"No task state found for {PRIVATE_STATE_KEY}: {task._private_state} "
                        f"(provenance: {task._provenance})",
                    )
                assert isinstance(task, Task)

                # remove any metadata and callbacks before changing the provenance
                self.remove_state(task)

                task._add_input_provenance(old_task)
                task._add_worker_provenance(graph_task)

                # then we can add it to the main graph using the right consumer
                assert graph_task.exit_worker is not None
                assert graph_task.exit_worker._graph is not None
                assert graph_task.exit_worker._graph._dispatcher is not None
                consumer = graph_task._get_consumer(task)
                graph_task.exit_worker._graph._dispatcher.add_work(consumer, task)

        instance = AdapterSinkWorker()
        self.graph.add_workers(instance)
        self.graph.set_dependency(self.exit_worker, instance)
        self.graph.set_entry(self.entry_worker)
        self.graph.finalize()  # compute the worker distances
        self.graph.init_workers()

    def get_task_classes(self) -> List[Type[Task]]:
        # usually the entry task gets dynamically determined from consume_work but we are overriding it here
        assert self.entry_worker is not None
        return self.entry_worker.get_task_classes()

    def init(self):
        # we need to install the graph dispatcher into the sub-graph
        assert self._graph is not None
        self.graph._dispatcher = self._graph._dispatcher

    def consume_work(self, task: Task):
        new_task = task.copy_public()

        # save the task provenance
        # xxx - we really just need to remember the provenance of the task
        old_task = task.model_copy(deep=True)
        new_task.add_private_state(PRIVATE_STATE_KEY, old_task)

        # artificially increase the provenance
        logging.debug("Adding additional provenance for %s", old_task._provenance)
        self._graph._provenance_tracker._add_provenance(old_task)

        # get any associated state and re-inject it
        state = self.get_state(task)
        metadata = state["metadata"]
        callback = state["callback"]

        def inject_state(provenance: ProvenanceChain):
            self.graph.watch(provenance, self)
            # We inject True to indicate that extra provenance is still associated with the task
            logging.debug("Injecting state for %s in %s", provenance, self.name)
            with self.lock:
                self._state[provenance] = old_task

        # and dispatch it to the sub-graph. this also sets the task provenance to InitialTaskWorker
        assert self.entry_worker is not None
        self.graph._add_work(
            self.entry_worker,
            new_task,
            metadata=metadata,
            status_callback=callback,
            provenance_callback=inject_state,
        )

    def notify(self, prefix: str):
        # we don't know how many tasks a subgraph will produce. we need to hold on to
        # the extra provenance as long as there is work for the initial provenance. this
        # function will get called when all the work for the prefix is completed and
        # then we can remove additional provenance here and clean up state
        self.graph.unwatch(prefix, self)
        with self.lock:
            if prefix not in self._state:
                raise ValueError(
                    f"Task {prefix} does not have any associated state: {self._state}"
                )
            task = self._state.pop(prefix)

        logging.info(
            "Subgraph completed. Removing provenance for %s in %s",
            prefix,
            self.name,
        )
        self._graph._provenance_tracker._remove_provenance(task, self)

    def abort_work(self, provenance: ProvenanceChain):
        # map the provenance to the sub-graph provenance
        need_to_abort = []
        with self.lock:
            for prefix, task in self._state.items():
                if task._provenance[: len(provenance)] == list(provenance):
                    need_to_abort.append((prefix, provenance))
        # abort the mapped provenance in our graph
        for prefix, provenance in need_to_abort:
            logging.info(
                "Aborting %s in %s (mapped from %s)", prefix, self.name, provenance
            )
            self.graph.abort_work(prefix)


def SubGraphWorker(
    *,
    graph: Graph,
    entry_worker: Optional[TaskWorker] = None,
    exit_worker: Optional[TaskWorker] = None,
    name: str = "SubGraphWorker",
) -> SubGraphWorkerInternal:
    """
    Factory function to create a SubGraphWorker that manages a subgraph within a larger PlanAI graph.
    If entry or exit worker are not provided, they will be inferred from the graph.
    The entry worker is the first worker in the subgraph that receives tasks, and the exit worker
    is the last worker that produces outputs. The exit worker must have exactly one output type.

    Parameters
    ----------
    name : str, optional
        Custom name for the SubGraphWorker class, defaults to "SubGraphWorker"
    graph : Graph
        The graph that will be run as part of this TaskWorker
    entry_worker : TaskWorker, optional
        The entry point worker of the graph that receives initial tasks
    exit_worker : TaskWorker, optional
        The exit point worker of the graph that produces final outputs
        Must have exactly one output type

    Returns
    -------
    SubGraphWorkerInternal
        A new instance of SubGraphWorker with the specified configuration

    Raises
    ------
    ValueError
        If the exit_worker has more than one output type
    """
    # Create a new class with the custom name
    CustomClass = type(name, (SubGraphWorkerInternal,), {})
    return CustomClass(graph=graph, entry_worker=entry_worker, exit_worker=exit_worker)


def main():  # pragma: no cover
    import argparse
    import random
    import time
    from typing import Type

    parser = argparse.ArgumentParser(description="Simple Graph Example")
    parser.add_argument(
        "--run-dashboard", action="store_true", help="Run the web dashboard"
    )
    args = parser.parse_args()

    # Define custom Task classes
    class Task1WorkItem(Task):
        data: str

    class Task2WorkItem(Task):
        processed_data: str

    class Task3WorkItem(Task):
        final_result: str

    # Define custom TaskWorker classes
    class Task1Worker(TaskWorker):
        output_types: List[Type[Task]] = [Task2WorkItem]

        def consume_work(self, task: Task1WorkItem):
            self.print(f"Task1 consuming: {task.data}")
            time.sleep(random.uniform(0.2, 0.9))
            for i in range(7):
                processed = f"Processed: {task.data.upper()} at iteration {i}"
                self.publish_work(
                    Task2WorkItem(processed_data=processed), input_task=task
                )

    class Task2Worker(TaskWorker):
        output_types: List[Type[Task]] = [Task3WorkItem]

        def consume_work(self, task: Task2WorkItem):
            self.print(f"Task2 consuming: {task.processed_data}")
            time.sleep(random.uniform(0.3, 2.5))

            if args.run_dashboard:
                # demonstrate the ability to request user input
                if random.random() < 0.15:
                    result, mime_type = self.request_user_input(
                        task=task,
                        instruction="Please provide a value",
                        accepted_mime_types=["text/html", "application/pdf"],
                    )
                    self.print(
                        f"User input: {len(result) if result else None} ({mime_type})"
                    )

            for i in range(11):
                final = f"Final: {task.processed_data} at iteration {i}!"
                self.publish_work(Task3WorkItem(final_result=final), input_task=task)

    class Task3Worker(TaskWorker):
        output_types: List[Type[Task]] = []

        def consume_work(self, task: Task3WorkItem):
            self.print(f"Task3 consuming: {task.final_result}")
            time.sleep(random.uniform(0.4, 1.2))
            self.print("Workflow complete!")

    # Create Graph
    sub_graph = Graph(name="Simple SubGraph")

    # Create tasks
    task1 = Task1Worker()
    task2 = Task2Worker()

    # Add tasks to Graph
    sub_graph.add_workers(task1, task2)

    # Set dependencies
    sub_graph.set_dependency(task1, task2)

    # Create the graph task
    graph_task = SubGraphWorker(
        name="SubGraph", graph=sub_graph, entry_worker=task1, exit_worker=task2
    )

    # Create the final consumer
    task3 = Task3Worker()

    graph = Graph(name="Simple Graph")
    graph.add_workers(graph_task, task3)
    graph.set_dependency(graph_task, task3)

    # Prepare initial work item
    initial_work = [
        (graph_task, Task1WorkItem(data="Hello, Graph v1!")),
        (graph_task, Task1WorkItem(data="Hello, Graph v2!")),
    ]

    # Run the Graph
    graph.run(
        initial_work,
        run_dashboard=args.run_dashboard,
        display_terminal=not args.run_dashboard,
    )


if __name__ == "__main__":
    main()



"""Provides chat-based interaction capabilities for PlanAI workflows.

This module enables the integration of traditional chat-based interactions with PlanAI's
complex graph-based workflows. The ChatTaskWorker allows developers to create interactive
sessions where users can:
- Engage with results from complex graph executions
- Trigger new workflow branches based on their inputs
- Maintain contextual conversations while leveraging PlanAI's advanced features

The chat functionality serves as a bridge between automated workflows and user-driven
processes, making it ideal for applications that require human oversight or
interactive decision-making within larger automated systems.
"""

# Copyright 2025 Niels Provos
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from datetime import datetime
from typing import List, Type

from .llm_task import BaseLLMTaskWorker
from .task import Task


class ChatMessage(Task):
    role: str
    content: str


class ChatTask(Task):
    messages: List[ChatMessage]


class ChatTaskWorker(BaseLLMTaskWorker):
    output_types: List[Type[Task]] = [ChatMessage]
    system_prompt: str = "You are a helpful AI assistant. Today is {date}."

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def _format_messages(self, messages: List[ChatMessage]):
        formatted_messages = [
            {
                "role": "system",
                "content": self.system_prompt.format(
                    date=datetime.today().strftime("%Y-%m-%d")
                ),
            },
        ]
        formatted_messages.extend(
            [{"role": message.role, "content": message.content} for message in messages]
        )
        return formatted_messages

    def consume_work(self, task: ChatTask):
        response = self.llm.chat(
            self._format_messages(task.messages),
            tools=self.tools if self.tools else None,
            allow_json_mode=False,
        )
        self.publish_work(ChatMessage(role="assistant", content=response), task)



from queue import Queue
from typing import Any, List, Optional, Tuple

from pydantic import BaseModel, Field, PrivateAttr

ProvenanceChain = Tuple[Tuple[str, int], ...]


class UserInputRequest(BaseModel):
    """
    Allows a PlanAI graph to ask for out-of-band user input.
    A consumer of a UserInputRequest needs to return a response
    via the respond() method.
    The request is a blocking call until the response is received.
    """

    task_id: str
    instruction: str
    provenance: Optional[ProvenanceChain] = None
    accepted_mime_types: List[str] = Field(
        default_factory=lambda: ["text/html", "application/pdf"]
    )
    _response_queue: Queue[Tuple[Any, Optional[str]]] = PrivateAttr(
        default_factory=Queue
    )

    def respond(self, response: Any, mime_type: Optional[str] = None) -> None:
        """
        Respond to the request with a response and an optional mime type.
        """
        if mime_type is not None and mime_type not in self.accepted_mime_types:
            raise ValueError(f"Mime type {mime_type} not accepted")

        self._response_queue.put((response, mime_type))



"""
Debug helper utilities for development and troubleshooting, particularly useful for multi-threaded applications.

This module provides debugging tools that help identify potential bugs and trace execution flow
in complex applications. The logging decorators are especially valuable when debugging issues
that arise from concurrent execution and race conditions in multi-threaded environments.

The utilities help track:
- Function call order and timing
- Caller context and source location
- Argument values passed between functions
- Execution flow across different threads

Usage:
    @log_function_call
    def my_function(arg1, arg2):
        # Function implementation
        pass
"""

import functools
import inspect
import logging


def log_function_call(func):
    """
    Decorator that logs the caller's information, function name, and arguments.
    """

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        # Get caller information
        frame = inspect.currentframe().f_back
        caller_name = "Unknown"  # Default if caller information is unavailable
        caller_line = 0

        if frame:
            caller_name = (
                frame.f_code.co_name
                if frame.f_code.co_name != "<module>"
                else "Module-level"
            )
            caller_line = frame.f_lineno

        # Get function name and arguments
        func_name = func.__name__
        arg_names = inspect.getfullargspec(func).args
        all_args = dict(zip(arg_names, args))
        all_args.update(kwargs)

        # Log the call information
        logging.info(
            f"Caller: {caller_name} (line {caller_line}) -> Function: {func_name}, Arguments: {all_args}"
        )

        # Call the original function
        return func(*args, **kwargs)

    return wrapper



import re
import tempfile
from textwrap import dedent
from typing import Callable, List, Optional, Tuple, Type

from llm_interface import LLMInterface
from pydantic import Field

from ..cached_task import CachedTaskWorker
from ..graph import Graph
from ..graph_task import SubGraphWorker
from ..integrations import SerperGoogleSearchTool, WebBrowser
from ..joined_task import JoinedTaskWorker
from ..llm_task import CachedLLMTaskWorker
from ..task import Task, TaskWorker


class SearchQuery(Task):
    query: str = Field(..., description="The search query to execute")
    metadata: Optional[str] = Field(
        None,
        description="Metadata to pass along with the query",
    )


class SearchResult(Task):
    title: str = Field(description="Title of the search result")
    link: str = Field(description="URL of the search result")
    snippet: Optional[str] = Field(
        default=None, description="Snippet of the search result"
    )


class SearchResults(Task):
    results: List[SearchResult] = Field(description="List of search results")


class PageResult(Task):
    url: str = Field(description="URL of the page")
    title: str = Field(description="Title of the page")
    content: Optional[str] = Field(default=None, description="Content of the page")


class PageAnalysis(Task):
    cleaned_content: Optional[str] = Field(
        None,
        description="Content after cleaning and filtering",
    )
    summary: str = Field(..., description="Summary of the analysis")
    is_relevant: bool = Field(
        ..., description="Whether the page contains relevant content"
    )


class ConsolidatedPages(Task):
    pages: List[PageResult] = Field(description="List of consolidated pages")


class SearchExecutor(CachedTaskWorker):
    output_types: List[Type[Task]] = [SearchResults]
    max_results: int = Field(10, description="Maximum number of results per query")

    def pre_consume_work(self, task: SearchQuery):
        self.notify_status(task, f"Searching for: {task.query}")
        self.print(f"Executing search for: {task.query}")

    def consume_work(self, task: SearchQuery):
        results = SerperGoogleSearchTool.search_internet(
            task.query, num_results=self.max_results, print_func=self.print
        )
        self.publish_work(
            task=SearchResults(
                results=[
                    SearchResult(title=r["title"], link=r["link"], snippet=r["snippet"])
                    for r in results
                ]
            ),
            input_task=task,
        )


class SearchResultSplitter(TaskWorker):
    """Splits search results into individual tasks for further processing.

    An optional filter function can be provided to exclude certain results.
    It should return True to include a result.
    """

    output_types: List[Type[Task]] = [SearchResult, ConsolidatedPages]
    filter_func: Optional[Callable[[SearchResult], bool]] = None

    def consume_work(self, task: SearchResults):
        results = task.results
        if self.filter_func:
            results = [r for r in results if self.filter_func(r)]
        if not results:
            # it's possible that all results were filtered out
            # we need to publish an empty ConsolidatedPages task to indicate failure
            self.publish_work(ConsolidatedPages(pages=[]), input_task=task)
            return
        for result in results:
            self.publish_work(task=result, input_task=task)


class PageFetcher(CachedTaskWorker):
    output_types: List[Type[Task]] = [PageResult, PageAnalysis]
    extract_pdf_func: Optional[Callable] = None
    support_user_input: bool = False

    def __init__(self, extract_pdf_func: Optional[Callable] = None):
        super().__init__()
        self.extract_pdf_func = extract_pdf_func

    def pre_consume_work(self, task):
        self.notify_status(task, f"Fetching content from: {task.link}")
        self.print(f"Fetching content from: {task.link}")

    def consume_work(self, task: SearchResult):
        content = WebBrowser.get_markdown_from_page(
            task.link,
            extract_markdown_from_pdf=self.extract_pdf_func,
            print_func=self.print,
        )

        if content is None and self.support_user_input:
            self.print(f"Failed to fetch content for {task.link}, requesting from user")
            data, mime_type = self.request_user_input(
                task=task,
                instruction=f'Please provide the content for: <a href="{task.link}" target="_blank">{task.link}</a>',
                accepted_mime_types=["text/html", "application/pdf"],
            )

            if data and mime_type:
                if mime_type == "application/pdf" and self.extract_pdf_func:
                    with tempfile.NamedTemporaryFile(delete=True) as temp_file:
                        temp_file.write(data)
                        temp_file.flush()
                        content = self.extract_pdf_func(temp_file.name, self.print)
                else:  # text/html
                    content = WebBrowser.extract_markdown(data)

        if content:
            # Remove markdown links while preserving the link text
            content = re.sub(r"\[([^\]]+)\]\([^)]+\)", r"\1", content)

            result = PageResult(url=task.link, title=task.title, content=content)
            self.publish_work(task=result, input_task=task)
        else:
            result = PageAnalysis(
                is_relevant=False,
                summary=f"Failed to fetch content from {task.link}",
                cleaned_content=None,
            )
            self.publish_work(task=result, input_task=task)


class PageRelevanceFilter(CachedLLMTaskWorker):
    output_types: List[Type[Task]] = [PageAnalysis]
    llm_input_type: Type[Task] = PageResult
    use_xml: bool = True
    debug_mode: bool = False
    prompt: str = dedent(
        """
You are provided with a block of page content that may include headers, footers, navigation links, advertisements, and other non-essential elements.
Your task is to extract and retain ONLY the information directly or indirectly relevant to the user's query and its associated metadata.
It is critical that all essential data—such as code snippets, numerical data, statistics, technical instructions, and contextual details—be preserved.

Input Details:
[query]
{query}
[/query]

[metadata]
{metadata}
[/metadata]

Follow these steps:

1. Thoroughly review the entire page content to identify all segments that explicitly or subtly address the query ([query]) and correlate with the metadata ([metadata]).

2. Extract every relevant detail, ensuring you capture:
   - Factual information, statistics, and data points
   - All code snippets, algorithms, and technical instructions
   - Contextual descriptions that enhance the understanding of the query

3. Remove non-essential elements such as headers, footers, navigation menus, and promotional material.
If any of these elements provide indirect yet significant context, include them with.

4. Cross-reference the extracted information with the provided metadata ([metadata]) to verify accuracy and completeness.

5. Consolidate the retained information into a clear and organized response using bullet points, numbered lists, or subheadings to enhance readability.

6. End your response with a brief summary explaining your extraction process and the reasoning behind the inclusion or exclusion of content.

If no relevant content is found, you must set is_relevant to False and cleaned_content to None.
        """
    ).strip()

    def format_prompt(self, task: PageResult):
        query = task.find_input_task(SearchQuery)
        if query is None:
            raise ValueError("SearchQuery not found in input tasks")
        return self.prompt.format(query=query.query, metadata=query.metadata)

    def pre_consume_work(self, task):
        self.notify_status(task, f"Analyzing relevance of: {task.url}")


class PageAnalysisConsumer(CachedTaskWorker):
    output_types: List[Type[Task]] = [PageResult]

    def consume_work(self, task: PageAnalysis):
        if task.is_relevant:
            result: PageResult = task.find_input_task(PageResult)
            if result is None:
                raise ValueError("PageAnalysisConsumer requires a PageResult input")
            result = result.copy_public()
            # inject the cleaned content into the result
            result.content = task.cleaned_content
            self.publish_work(task=result, input_task=task)
        else:
            result: SearchResult = task.find_input_task(SearchResult)
            if result is None:
                raise ValueError("PageAnalysisConsumer requires a SearchResult input")
            self.publish_work(
                PageResult(url=result.link, title=result.title),
                input_task=task,
            )


class PageConsolidator(JoinedTaskWorker):
    output_types: List[Type[Task]] = [ConsolidatedPages]
    join_type: Type[TaskWorker] = SearchExecutor

    def consume_work_joined(self, tasks: List[PageResult]):
        self.publish_work(
            task=ConsolidatedPages(pages=[t for t in tasks if t.content]),
            input_task=tasks[0],
        )


class FinalCollector(TaskWorker):
    output_types: List[Type[Task]] = [ConsolidatedPages]

    def consume_work(self, task: ConsolidatedPages):
        self.publish_work(task=task.copy_public(), input_task=task)


def create_search_fetch_graph(
    *,
    llm: LLMInterface,
    name: str = "SearchFetchWorker",
    filter_func: Optional[Callable[[SearchResult], bool]] = None,
    extract_pdf_func: Optional[Callable] = None,
) -> Tuple[Graph, TaskWorker, TaskWorker]:
    graph = Graph(name=f"{name}Graph", strict=True)

    search = SearchExecutor()
    splitter = SearchResultSplitter(filter_func=filter_func)
    fetcher = PageFetcher(extract_pdf_func=extract_pdf_func)
    relevance = PageRelevanceFilter(llm=llm)
    analysis_consumer = PageAnalysisConsumer()
    consolidator = PageConsolidator()
    final_collector = FinalCollector()

    graph.add_workers(
        search,
        splitter,
        fetcher,
        relevance,
        analysis_consumer,
        consolidator,
        final_collector,
    )
    graph.set_dependency(search, splitter).next(fetcher).next(relevance).next(
        analysis_consumer
    ).next(consolidator).next(final_collector)
    # if we can't fetch the content, we will bypass the relevance filter
    graph.set_dependency(fetcher, analysis_consumer)
    # if we filter out all results, we will bypass the consolidator
    graph.set_dependency(splitter, final_collector)

    return graph, search, final_collector


def create_search_fetch_worker(
    *,
    llm: LLMInterface,
    name: str = "SearchFetchWorker",
    filter_func: Optional[Callable[[SearchResult], bool]] = None,
    extract_pdf_func: Optional[Callable] = None,
) -> TaskWorker:
    """Creates a SubGraphWorker that searches and fetches web content.

    This worker creates a subgraph that processes a search query through multiple stages:
    1. Executes web search
    2. Fetches content from result pages
    3. Filters for relevance using LLM
    4. Consolidates relevant pages

    Args:
        llm: LLM interface for content analysis
        name: Name for the worker
        filter_func: Optional function to filter search results - should return True to include a result
        extract_pdf_func: Optional function to extract text from PDFs

    Input Task:
        SearchQuery: A task containing a single 'query' string to search for

    Output Task:
        ConsolidatedPages: A task containing a list of PageResult objects, each with:
            - url: The page URL
            - title: Page title
            - content: Extracted page content (if successfully fetched)

        In case of failure, the ConsolidatedPages task will contain an empty list.

    Returns:
        A SubGraphWorker that implements the search and fetch pattern
    """
    graph, search, consolidator = create_search_fetch_graph(
        llm=llm, name=name, filter_func=filter_func, extract_pdf_func=extract_pdf_func
    )

    return SubGraphWorker(
        name=name,
        graph=graph,
        entry_worker=search,
        exit_worker=consolidator,
    )



import sys

from diskcache import Cache


def handle_cache_subcommand(parsed_args):
    cache = Cache(parsed_args.cache_dir)
    if parsed_args.search_dirs:
        for search_dir in parsed_args.search_dirs:
            sys.path.append(search_dir)

    if parsed_args.clear:
        cache.clear()
        print("Cache cleared successfully.")
    elif parsed_args.delete:
        to_delete = []
        for key in cache:
            try:
                value, input_data = cache.get(key)
            except Exception as e:
                print(f"Error handling cache: {str(e)}", file=sys.stderr)
                continue

            if input_data.name == parsed_args.delete:
                to_delete.append(key)
        print(f"Deleting {len(to_delete)} entries.")
        for key in to_delete:
            del cache[key]
    else:
        total_size = 0
        print(f"Reading cache from: {parsed_args.cache_dir}")
        print("-----------------------------")

        if len(cache) == 0:
            print("The cache is empty.")
        else:
            # Dictionary to hold the statistics
            stats = {}

            for key in cache:
                try:
                    value, input_data = cache.get(key)
                except Exception as e:
                    print(f"Error handling cache: {str(e)}", file=sys.stderr)
                    continue

                input_name = input_data.name

                if parsed_args.output_task_filter:
                    if not any(
                        [task.name == parsed_args.output_task_filter for task in value]
                    ):
                        continue

                # Update statistics
                if input_name not in stats:
                    stats[input_name] = {"count": 0, "total_size": 0}

                stats[input_name]["count"] += 1
                # Estimate size of cached entry
                stats[input_name]["total_size"] += sum(sys.getsizeof(v) for v in value)

            # Print the statistics
            for input_name, data in stats.items():
                print(f"Input: {input_name}")
                print(f"  Number of Entries: {data['count']}")
                print(f"  Total Size (bytes): {data['total_size']}")
                print("-----------------------------")

                total_size += data["total_size"]
            print(f"Total Size of Cache (bytes): {total_size}")



# Copyright 2024 Niels Provos
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import logging
import shutil
import time
from collections import deque
from threading import Event, Thread
from typing import Any, Callable, Dict, List, Optional, Sequence, Set, Tuple, Type

from colorama import Fore, Style, init
from pydantic import BaseModel, ConfigDict, Field, PrivateAttr

from .dispatcher import Dispatcher
from .joined_task import InitialTaskWorker
from .provenance import ProvenanceChain, ProvenanceTracker
from .task import Task, TaskStatusCallback, TaskWorker
from .user_input import UserInputRequest

# Initialize colorama for Windows compatibility
init(autoreset=True)

# Provide the provenance for state tracking before work is added to the dispatcher
ProvenanceCallback = Callable[["ProvenanceChain"], None]


class Graph(BaseModel):
    """A graph for orchestrating task workers and their dependencies.

    The Graph class manages the execution flow of tasks through workers, handling dependencies,
    parallel execution, monitoring, and output collection. It supports both terminal-based
    and web dashboard monitoring of task execution.

    Attributes:
        name (str): Name identifier for the graph instance
        strict (bool): If True, the graph will enforce strict validation of tasks provided to publish_work()
        workers (Set[TaskWorker]): Set of task workers in the graph
        dependencies (Dict[TaskWorker, List[TaskWorker]]): Maps workers to their downstream dependencies

    Example:
        >>> graph = Graph(name="Data Processing Pipeline")
        >>> worker1 = DataLoader()
        >>> worker2 = DataProcessor()
        >>> graph.add_workers(worker1, worker2)
        >>> graph.set_dependency(worker1, worker2)
        >>> graph.run([(worker1, LoadTask(file="data.csv"))])
    """

    model_config = ConfigDict(arbitrary_types_allowed=True)

    name: str
    strict: bool = False
    workers: Set[TaskWorker] = Field(default_factory=set)
    dependencies: Dict[TaskWorker, List[TaskWorker]] = Field(default_factory=dict)

    _dispatcher: Optional[Dispatcher] = PrivateAttr(default=None)
    _dispatcher_created: bool = PrivateAttr(default=False)
    _terminal_thread: Optional[Thread] = PrivateAttr(default=None)
    _provenance_tracker: ProvenanceTracker = PrivateAttr(
        default_factory=ProvenanceTracker
    )

    _max_parallel_tasks: Dict[Type[TaskWorker], int] = PrivateAttr(default_factory=dict)
    _sink_tasks: List[Type[Task]] = PrivateAttr(default_factory=list)
    _sink_workers: List[TaskWorker] = PrivateAttr(default_factory=list)
    _exit_worker: Optional[TaskWorker] = PrivateAttr(default=None)
    _initial_worker: Optional[TaskWorker] = PrivateAttr(default=None)
    _subgraph_workers: Set[TaskWorker] = PrivateAttr(default_factory=set)

    _worker_distances: Dict[str, Dict[str, int]] = PrivateAttr(default_factory=dict)
    _has_terminal: bool = PrivateAttr(default=False)

    _user_request_callback: Optional[
        Callable[[Dict[str, Any], UserInputRequest], None]
    ] = PrivateAttr(default=None)

    def __init__(self, **data):
        super().__init__(**data)
        self._initial_worker = InitialTaskWorker()
        self._provenance_tracker = ProvenanceTracker(name=self.name)
        self.add_worker(self._initial_worker)

    def trace(self, prefix: ProvenanceChain):
        self._provenance_tracker.trace(prefix)

    def watch(
        self,
        prefix: ProvenanceChain,
        notifier: TaskWorker,
        callback: Optional[Callable[[ProvenanceChain], None]] = None,
    ) -> bool:
        return self._provenance_tracker.watch(prefix, notifier, callback)

    def unwatch(
        self,
        prefix: ProvenanceChain,
        notifier: TaskWorker,
        callback: Optional[Callable[[ProvenanceChain], None]] = None,
    ) -> bool:
        return self._provenance_tracker.unwatch(prefix, notifier, callback)

    def add_worker(self, worker: TaskWorker) -> "Graph":
        """Adds a single task worker to the graph.

        Args:
            task (TaskWorker): The worker instance to add to the graph

        Returns:
            Graph: The graph instance for method chaining

        Raises:
            ValueError: If the worker is already present in the graph

        Example:
            >>> graph = Graph(name="Pipeline")
            >>> worker = DataProcessor()
            >>> graph.add_worker(worker)
        """
        if worker in self.workers:
            raise ValueError(f"Task {worker} already exists in the Graph.")
        if self.strict:
            # This causes the worker to check during publish_work() that the provenance fields of the new
            # task are empty. This can help with a common bug patterns where a task is reused.
            worker._strict_checking = True

        self.workers.add(worker)
        # Check if any class in the MRO hierarchy is named 'SubGraphWorkerInternal'
        if any(
            cls.__name__ == "SubGraphWorkerInternal" for cls in type(worker).__mro__
        ):
            self._subgraph_workers.add(worker)
        self.dependencies[worker] = []
        worker.set_graph(self)
        return self

    def get_dispatcher(self) -> Optional[Dispatcher]:
        """Get the dispatcher instance for the graph."""
        return self._dispatcher

    def add_workers(self, *workers: TaskWorker) -> "Graph":
        """Add multiple tasks to the Graph."""
        for worker in workers:
            self.add_worker(worker)
        return self

    def get_worker_by_input_type(self, input_type: Type[Task]) -> Optional[TaskWorker]:
        """Get a worker that consumes a specific input type.

        This method searches through registered workers to find one that processes
        the specified input task type.

        Args:
            input_type (Type[Task]): The input task type class to match against workers.

        Returns:
            Optional[TaskWorker]: The matching worker if found, None otherwise.

        Example:
            worker = graph.get_worker_by_input_type(ImageTask)
        """
        for worker in self.workers:
            if input_type in worker.get_task_classes():
                return worker
        return None

    def get_worker_by_output_type(
        self, output_type: Type[Task]
    ) -> Optional[TaskWorker]:
        """Get a worker that produces a specific output type.

        This method searches through registered workers to find one that produces
        the specified output task type.

        Args:
            output_type (Type[Task]): The output task type class to match against workers.

        Returns:
            Optional[TaskWorker]: The matching worker if found, None otherwise.

        Example:
            worker = graph.get_worker_by_output_type(ImageTask)
        """
        for worker in self.workers:
            if output_type in worker.output_types:
                return worker
        return None

    def set_dependency(
        self, upstream: TaskWorker, downstream: TaskWorker
    ) -> TaskWorker:
        """Set a dependency between two tasks."""
        return self._set_dependency(upstream, downstream)

    def _set_dependency(
        self, upstream: TaskWorker, downstream: TaskWorker, register: bool = True
    ) -> TaskWorker:
        if upstream not in self.workers or downstream not in self.workers:
            raise ValueError(
                f"Both workers (upstream: {upstream.__class__.__name__}) (downstream: {downstream.__class__.__name__}) must be added to the Graph before setting dependencies."
            )

        if downstream not in self.dependencies[upstream]:
            self.dependencies[upstream].append(downstream)
            if register:
                potential_task_classes: List[Type[Task]] = downstream.get_task_classes()
                potential_output_types: List[Type[Task]] = upstream.output_types
                # intersect the two lists
                task_classes = set(potential_task_classes).intersection(
                    potential_output_types
                )
                if not task_classes:
                    raise ValueError(
                        f"Worker {upstream.name} does not produce any task types that worker {downstream.name} can consume."
                    )
                if len(task_classes) > 1:
                    raise ValueError(
                        f"Worker {upstream.name} produces multiple task types that worker {downstream.name} can consume. Please specify the task type."
                    )
                output_type = task_classes.pop()
                upstream.register_consumer(
                    task_cls=output_type,
                    consumer=downstream,
                )

        return downstream

    def set_sink(
        self,
        worker: TaskWorker,
        output_type: Type[Task],
        notify: Optional[Callable[[Dict[str, Any], Task], None]] = None,
    ) -> None:
        """Designates a worker as a data sink for collecting specific output tasks.

        A sink worker is a special endpoint in the graph that collects and optionally processes
        output tasks of a specific type. The sink can either store tasks for later retrieval
        or forward them to a notification callback.

        Args:
            worker (TaskWorker): The worker whose output should be collected
            output_type (Type[Task]): The specific task type to collect at this sink
            notify (Callable[[Dict[str, Any], Task], optional): Callback function
                that receives the task's metadata and the task itself. If provided, tasks
                won't be stored in the sink's collection.

        Raises:
            ValueError: If the specified worker doesn't have the correct output type.
            ValueError: If the specifier output type has already been registered as an output

        Note:
            - The sink worker is automatically added to the graph and set as a dependency of the specified worker.
            - The output type of the specified worker is used to type the tasks consumed by the sink.
            - Only workers with a single output type can be set as sinks.

        Example:
            >>> graph = Graph(name="Example Graph")
            >>> worker = SomeTaskWorker()
            >>> graph.add_worker(worker)
            >>> graph.set_sink(worker, OutputTask)
            >>> graph.run(initial_tasks=[(worker, SomeTask())])
            >>> results = graph.get_output_tasks()
        """
        if output_type not in worker.output_types:
            raise ValueError(
                f"Worker {worker.name} does not have output type {output_type.__name__} to use for a sink"
            )

        class SinkWorker(TaskWorker):
            def __init__(self, graph: "Graph", **data):
                super().__init__(**data)
                self._graph = graph
                self._notify = notify

            def consume_work(self, task: output_type):
                if self._notify:
                    metadata = self.get_metadata(task)
                    logging.info("SinkWorker is notifying on task %s", task.name)
                    self._notify(metadata, task)
                else:
                    with self.lock:
                        self._graph._sink_tasks.append(task)
                # we should not remove metadata or callbacks here as we don't know
                # that all provenance has been removed

        # Create a new class with the specific output type
        instance = SinkWorker(self)
        self.add_worker(instance)
        self.set_dependency(worker, instance)

        # Set this as the sink worker for the graph
        self._sink_workers.append(instance)
        logging.info(
            "Sink on worker %s set for output type %s",
            worker.name,
            output_type.__name__,
        )

    def get_output_tasks(self) -> List[Type[Task]]:
        """
        Retrieves all tasks that were consumed by the sink workers in the graph.

        This method returns a list of tasks that were collected by all sink workers
        after the graph has been run. Each task in the list is an instance of the
        output type specified when the corresponding sink was set.

        Returns:
            List[Type[Task]]: A list of tasks collected by all sink workers. The actual
            type of each task depends on the output types of the workers set as sinks.

        Note:
            - This method should be called after the graph has been run using the `run()` method.
            - If no sinks were set or if the graph hasn't been run, this method will return an empty list.
            - The order of tasks in the list corresponds to the order they were consumed by the sink workers.

        Example:
            >>> graph = Graph(name="Example Graph")
            >>> worker = SomeTaskWorker()
            >>> graph.add_worker(worker)
            >>> graph.set_sink(worker)
            >>> graph.run(initial_tasks=[(worker, SomeTask())])
            >>> results = graph.get_output_tasks()

        See Also:
            set_sink(): Method for setting a worker as a sink in the graph.
        """
        return self._sink_tasks

    def set_max_parallel_tasks(
        self, worker_class: Type[TaskWorker], max_parallel_tasks: int
    ) -> None:
        """
        Set the maximum number of parallel tasks for a specific worker class.

        Args:
            worker_class (Type[TaskWorker]): The class of the worker to limit.
            max_parallel_tasks (int): The maximum number of parallel tasks allowed.

        Note:
            This setting will be applied to the dispatcher when the graph is run.
            If the dispatcher is already running, it will update the limit dynamically.
        """
        if not issubclass(worker_class, TaskWorker):
            raise ValueError(f"{worker_class.__name__} is not a subclass of TaskWorker")

        if max_parallel_tasks <= 0:
            raise ValueError("max_parallel_tasks must be greater than 0")

        # Store the setting
        self._max_parallel_tasks[worker_class] = max_parallel_tasks

        # If dispatcher is already running, update it
        if self._dispatcher:
            self._dispatcher.set_max_parallel_tasks(worker_class, max_parallel_tasks)

    def validate_graph(self) -> None:
        """Return the execution order of tasks based on dependencies."""
        pass

    def finalize(self):
        self.compute_worker_distances()

    def prepare(
        self,
        run_dashboard: bool = False,
        display_terminal: bool = True,
        dashboard_port: int = 5000,
    ) -> None:
        """Initializes the graph for execution by setting up monitoring and worker components.

        This method must be called before executing tasks. It sets up:
        - Task dispatcher for managing worker execution
        - Optional web dashboard for monitoring
        - Optional terminal-based status display
        - Worker parallel execution limits

        Args:
            run_dashboard (bool): If True, starts a web interface for monitoring
            display_terminal (bool): If True, shows execution progress in terminal
            dashboard_port (int): Port number for the web dashboard if enabled

        Example:
            >>> graph = Graph(name="Pipeline")
            >>> # ... add workers and dependencies ...
            >>> graph.prepare(run_dashboard=True, dashboard_port=8080)
            >>> graph.execute(initial_tasks)
        """
        # Empty the sink tasks
        if self._sink_workers:
            self._sink_tasks = []
            if run_dashboard:
                logging.warning(
                    "The dashboard will make the graph wait for manual termination. "
                    "This is usually not desired when using a sink worker."
                )

        # Allow workers to log messages
        self._log_lines = []

        # Start the dispatcher
        dispatcher = self._dispatcher
        if self._dispatcher is None:
            dispatcher = Dispatcher(self, web_port=dashboard_port)
            self._dispatcher_created = True
        elif not self._dispatcher_created:
            logging.info("Graph %s is using an existing dispatcher", self.name)
            if run_dashboard or display_terminal:
                raise RuntimeError(
                    "Dispatcher is already running. Should not start dashboard or terminal display."
                )

        dispatcher.start()
        if run_dashboard:
            dispatcher.start_web_interface()
        self._dispatcher = dispatcher

        if display_terminal:
            self._has_terminal = True
            self._start_terminal_display()
            self._terminal_thread = Thread(target=self._terminal_display_thread)
            self._terminal_thread.start()
        else:
            self._has_terminal = False

        # Apply the max parallel tasks settings
        for worker_class, max_parallel_tasks in self._max_parallel_tasks.items():
            dispatcher.set_max_parallel_tasks(worker_class, max_parallel_tasks)

    def run(
        self,
        initial_tasks: Sequence[Tuple[TaskWorker, Task]],
        run_dashboard: bool = False,
        display_terminal: bool = True,
        dashboard_port: int = 5000,
    ) -> None:
        """
        Execute the Graph by initiating source tasks and managing the workflow.

        This method sets up the Dispatcher, initializes workers, and processes the initial tasks.
        It manages the entire execution flow of the graph until completion.

        Args:
            initial_tasks (List[Tuple[TaskWorker, Task]]): A list of tuples, each containing
                a TaskWorker and its corresponding Task to initiate the graph execution.
            run_dashboard (bool, optional): If True, starts a web interface for monitoring the
                graph execution. Defaults to False.
            display_terminal (bool, optional): If True, displays a terminal status for the graph
                execution. Defaults to True.
            dashboard_port (int, optional): The port number for the web interface. Defaults to 5000.

        Raises:
            ValueError: If any of the initial tasks fail validation.

        Note:
            - This method blocks until all tasks in the graph are completed.
            - It handles the initialization and cleanup of workers, dispatcher, and thread pool.
            - If run_dashboard is True, the method will wait for manual termination of the web interface.

        Example:
            graph = Graph(name="My Workflow")
            # ... (add workers and set dependencies)
            initial_work = [(task1, Task1WorkItem(data="Start"))]
            graph.run(initial_work, run_dashboard=True)
        """
        self.prepare(
            run_dashboard=run_dashboard,
            display_terminal=display_terminal,
            dashboard_port=dashboard_port,
        )
        self.execute(initial_tasks)

    def execute(self, initial_tasks: Sequence[Tuple[TaskWorker, Task]]) -> None:
        """Executes the graph with the provided initial tasks.

        This method starts the actual task processing in the graph. It should be called
        after prepare() has been used to set up the execution environment.

        Args:
            initial_tasks (Sequence[Tuple[TaskWorker, Task]]): A sequence of worker-task
                pairs to start the graph execution

        Raises:
            Exception: If task validation fails for any worker-task pair
            RuntimeError: If prepare() hasn't been called first

        Note:
            - Blocks until all tasks complete unless a dashboard is running
            - Automatically handles worker initialization and cleanup
            - Maintains execution state for monitoring and debugging

        Example:
            >>> graph.prepare()
            >>> initial = [(worker, Task(data="start"))]
            >>> graph.execute(initial)
        """
        self.set_entry(*[x[0] for x in initial_tasks])
        # Finalize the graph (compute distances)
        self.finalize()

        # Check special needs for all workers
        for worker in self.workers:
            if hasattr(worker, "_validate_post_register") and callable(
                worker._validate_post_register
            ):
                worker._validate_post_register()

        # let the workers know that we are about to start
        self.init_workers()

        for worker, task in initial_tasks:
            success, error = worker.validate_task(type(task), worker)
            if not success:
                raise error

            self._add_work(worker, task)

        # Wait for all tasks to complete
        logging.info("Graph %s started - waiting for completion", self.name)
        self._dispatcher.wait_for_completion(
            wait_for_quit=self._dispatcher.has_dashboard()
        )
        logging.info("Graph %s completed", self.name)

        # we'll stop the dispatcher only if we created it
        if self._dispatcher_created:
            self._dispatcher.stop()
            logging.info("Dispatcher stopped")

        if self._has_terminal:
            self._stop_terminal_display_event.set()
            self._terminal_thread.join()
            logging.info("Terminal display stopped")

        for worker in self.workers:
            worker.completed()

        logging.info("All workers completed - execution finished")

    def init_workers(self):
        for worker in self.workers:
            worker.init()

    def _add_work(
        self,
        worker: TaskWorker,
        task: Task,
        metadata: Optional[Dict] = None,
        status_callback: Optional[TaskStatusCallback] = None,
        provenance_callback: Optional[ProvenanceCallback] = None,
    ) -> ProvenanceChain:
        provenance = self._initial_worker.get_next_provenance()
        task._provenance = [provenance] + task._provenance

        prov_chain = task.prefix(1)

        # Register state before adding work
        if metadata or status_callback:
            self._provenance_tracker.add_state(prov_chain, metadata, status_callback)

        # Allows workers to do additional state tracking based on the provenance
        if provenance_callback:
            provenance_callback(prov_chain)

        assert self._dispatcher is not None
        self._dispatcher.add_work(worker, task)
        return prov_chain

    def add_work(
        self,
        worker: TaskWorker,
        task: Task,
        metadata: Optional[Dict] = None,
        status_callback: Optional[TaskStatusCallback] = None,
    ) -> ProvenanceChain:
        if worker not in self.dependencies[self._initial_worker]:
            raise ValueError(
                f"Worker {worker.name} is not an entry point to the Graph."
            )
        return self._add_work(worker, task, metadata, status_callback)

    def abort_work(self, provenance: ProvenanceChain) -> bool:
        """
        This method attempts to abort work that is currently in progress. It first checks if the
        provenance chain exists in the provenance tracker. If found, it aborts the work through
        the dispatcher and propagates the abort request to all subgraph workers. If the provenance
        chain is not found, a warning is logged.

        Returns:
            bool: True if the work was aborted successfully, False otherwise
        """
        # We don't need to worry about race conditions here as provenance chains are unique
        # and either they exists in the provenance tracker or they have completed already
        if self._provenance_tracker.has_provenance(provenance):
            self._dispatcher.abort_work(self, provenance)
            for worker in self._subgraph_workers:
                worker.abort_work(provenance)
            return True

        logging.warning("Provenance %s not found to abort work", provenance)
        return False

    def set_entry(self, *workers: TaskWorker) -> "Graph":
        """Set the workers that are entry points to the Graph.

        This method establishes connections from the initial (root) worker to the specified
        workers, marking them as entry points in the execution graph.

        Args:
            *workers (TaskWorker): Variable number of TaskWorker instances to be set as
                entry points.

        Returns:
            Graph: The Graph instance itself for method chaining.

        Example:
            ```
            graph = Graph()
            worker1 = TaskWorker()
            worker2 = TaskWorker()
            graph.set_entry(worker1, worker2)
            ```
        """
        for worker in workers:
            self._set_dependency(self._initial_worker, worker, register=False)
        return self

    def get_entry_workers(self) -> List[TaskWorker]:
        """Get the entry workers of the graph.

        Returns:
            List[TaskWorker]: A list of TaskWorker instances that are entry points in the graph.

        Example:
            ```
            entry_workers = graph.get_entry_workers()
            ```
        """
        assert self._initial_worker is not None
        return self.dependencies.get(self._initial_worker, [])

    def set_exit(self, worker: TaskWorker) -> None:
        """Set the exit worker for the graph.

        This method designates a specific worker as the exit point of the graph. This is useful
        for simplifying the creation of subgraphs and provides a programmatic way to define
        the end of the workflow.

        Args:
            worker (TaskWorker): The TaskWorker instance to be set as the exit worker.

        Example:
            ```
            graph.set_exit_worker(worker)
            ```
        """
        if worker not in self.workers:
            raise ValueError(
                f"Worker {worker.name} must be added to the Graph before setting it as an exit worker."
            )
        self._exit_worker = worker

    def get_exit_worker(self) -> Optional[TaskWorker]:
        """Get the exit worker of the graph.

        Returns:
            Optional[TaskWorker]: The TaskWorker instance that is set as the exit worker, or None if not set.

        Example:
            ```
            exit_worker = graph.get_exit_worker()
            ```
        """
        return self._exit_worker

    def compute_worker_distances(self):
        for worker in self.workers:
            self._worker_distances[worker.name] = self._bfs_distances(worker)

    def _bfs_distances(self, start_worker: TaskWorker) -> Dict[str, int]:
        distances = {start_worker.name: 0}
        queue = deque([(start_worker, 0)])
        visited = set()

        while queue:
            current_worker, distance = queue.popleft()
            if current_worker in visited:
                continue
            visited.add(current_worker)

            for downstream in self.dependencies.get(current_worker, []):
                if (
                    downstream.name not in distances
                    or distance + 1 < distances[downstream.name]
                ):
                    distances[downstream.name] = distance + 1
                    queue.append((downstream, distance + 1))

        return distances

    def _start_terminal_display(self):
        self._stop_terminal_display_event = Event()
        self._stop_terminal_display_event.clear()

    def _terminal_display_thread(self):
        try:
            while not self._stop_terminal_display_event.is_set():
                self.display_terminal_status()
                time.sleep(0.25)  # Update interval
        finally:
            self._clear_terminal()
            self._print_log()

    def _clear_terminal(self):
        # Clear the terminal when the thread is terminating
        print("\033[H\033[J")

    def _print_log(self, max_lines=15):
        print("\nLog:")
        if max_lines <= 0:
            for line in self._log_lines:
                print(line)
            return

        # Get terminal width and leave a margin of 10 characters
        terminal_width = max(shutil.get_terminal_size().columns - 10, 20)

        # Flatten and wrap the log lines
        flattened_lines = []
        for line in self._log_lines:
            for subline in line.splitlines():
                while len(subline) > terminal_width:
                    flattened_lines.append(subline[:terminal_width])
                    subline = subline[terminal_width:]
                flattened_lines.append(subline)

        # Print the last 'max_lines' number of lines
        for line in flattened_lines[-max_lines:]:
            print(line)

    def display_terminal_status(self):
        stats = self._dispatcher.get_execution_statistics()
        terminal_size = shutil.get_terminal_size((80, 20))
        terminal_width = terminal_size.columns
        terminal_height = terminal_size.lines

        print("\033[H\033[J")  # Clear terminal

        # Sort the workers based on their distance from InitialTaskWorker
        distances = self._worker_distances.get("InitialTaskWorker", {})
        sorted_workers = sorted(
            stats.items(), key=lambda item: distances.get(item[0], float("inf"))
        )

        for worker, data in sorted_workers:
            completed = data["completed"]
            active = data["active"]
            queued = data["queued"]
            failed = data["failed"]

            total_tasks = completed + active + queued + failed
            available_width = (
                terminal_width - 60
            ) // 2  # Adjust for worker name and separator

            if total_tasks > available_width:
                scale_factor = available_width / total_tasks
                completed_scaled = (
                    max(1, int(completed * scale_factor)) if completed > 0 else 0
                )
                active_scaled = max(1, int(active * scale_factor)) if active > 0 else 0
                queued_scaled = max(1, int(queued * scale_factor)) if queued > 0 else 0
                failed_scaled = max(1, int(failed * scale_factor)) if failed > 0 else 0
            else:
                completed_scaled, active_scaled, queued_scaled, failed_scaled = (
                    completed,
                    active,
                    queued,
                    failed,
                )

            # Create bars
            completed_bar = Fore.GREEN + "🟩" * completed_scaled
            active_bar = Fore.BLUE + "🔵" * active_scaled
            queued_bar = Fore.LIGHTYELLOW_EX + "🟠" * queued_scaled
            failed_bar = Fore.RED + "❌" * failed_scaled

            # First print: worker name and bars
            status_line = f"{worker[:25]:25} | {completed_bar}{active_bar}{queued_bar}{failed_bar}{Style.RESET_ALL}"
            print(status_line, end="")

            # Calculate and print counts at the right edge
            counts = f"C:{completed} A:{active} Q:{queued} F:{failed}"

            # Move cursor to the right edge minus the length of counts
            cursor_move = f"\033[{terminal_width - len(counts)}G"

            print(f"{cursor_move}{counts}")

        lines_to_print = terminal_height - len(sorted_workers) - 4
        self._print_log(max_lines=lines_to_print)

        # Reset the cursor to the top
        print("\033[H")

    def print(self, *args):
        message = " ".join(str(arg) for arg in args)
        logging.info("Application: %s", message)
        if self._dispatcher is not None and self._dispatcher.has_dashboard():
            self._dispatcher.add_log(message)
        if self._has_terminal:
            self._log_lines.append(message)
        else:
            print(*args)

    def __str__(self) -> str:
        return f"Graph: {self.name} with {len(self.workers)} tasks"

    def __repr__(self) -> str:
        return self.__str__()

    def shutdown(self, timeout: float = 5.0) -> bool:
        """Gracefully shuts down the graph and all its components.

        Args:
            timeout (float): Maximum time to wait for tasks to complete in seconds

        Returns:
            bool: True if shutdown was successful, False if timeout occurred
        """
        logging.info("Initiating graph shutdown...")

        if not self._dispatcher:
            # nothing to do if the dispatcher was never created
            return True

        if self._dispatcher_created:
            # Signal dispatcher to stop accepting new work
            self._dispatcher.initiate_shutdown()

            # Wait for active tasks to complete or timeout
            self._dispatcher.wait_for_completion(timeout=timeout, wait_for_quit=False)

            # Stop the dispatcher thread
            self._dispatcher.stop(timeout=1.0)

            # Stop terminal display if active
            if self._has_terminal and self._terminal_thread:
                self._stop_terminal_display_event.set()
                self._terminal_thread.join(timeout=1.0)
        else:
            self._dispatcher.deregister_graph(self)

        logging.info("Graph shutdown completed successfully")
        return True

    def register_dispatcher(self, dispatcher: Dispatcher) -> None:
        """Register a dispatcher for this graph.

        Registers a dispatcher instance with the graph. This is useful when running multiple different graphs.

        Args:
            dispatcher (Dispatcher): The dispatcher instance to register.

        Returns:
            None

        """
        self._dispatcher = dispatcher
        self._dispatcher_created = False
        self._dispatcher.register_graph(self)

    def set_user_request_callback(
        self, callback: Callable[[Dict[str, Any], UserInputRequest], None]
    ) -> None:
        """
        Set a callback function to handle user input requests. The callback will receive task associated
        metadata and the actual UserInputRequest object.
        """
        self._user_request_callback = callback

    def wait_on_user_request(
        self, request: UserInputRequest
    ) -> Tuple[Any, Optional[str]]:
        """Wait for user input request to be completed.

        Args:
            request (UserInputRequest): The user input request to wait for. Blocks the thread until
              the user provides input.

        Returns:
            Tuple[Any, Optional[str]]: A tuple where the first element is the user's input (data), and the second
              element (if available) is the MIME type of the provided data.
        """

        if self._user_request_callback:
            assert request.provenance
            prefix = request.provenance[:1]
            metadata = {}
            state = self._provenance_tracker.get_state(prefix)
            if state and "metadata" in state:
                metadata = state["metadata"]
            self._user_request_callback(metadata, request)
        else:
            # the default path is to use the dispatcher dashboard
            assert self._dispatcher is not None
            queue = self._dispatcher.get_user_input_requests_queue()
            queue.put(request)

        # in any case, we expect the consumer of user request to put the result in this queue
        result, mime_type = (
            request._response_queue.get()
        )  # Block until result is available

        return result, mime_type


def main():  # pragma: no cover
    import argparse
    import random

    parser = argparse.ArgumentParser(description="Simple Graph Example")
    parser.add_argument(
        "--run-dashboard", action="store_true", help="Run the web dashboard"
    )
    args = parser.parse_args()

    # Define custom Task classes
    class Task1WorkItem(Task):
        data: str

    class Task2WorkItem(Task):
        processed_data: str

    class Task3WorkItem(Task):
        final_result: str

    # Define custom TaskWorker classes
    class Task1Worker(TaskWorker):
        output_types: List[Type[Task]] = [Task2WorkItem]

        def consume_work(self, task: Task1WorkItem):
            self.print(f"Task1 consuming: {task.data}")
            time.sleep(random.uniform(0.4, 1.5))
            for i in range(7):
                processed = f"Processed: {task.data.upper()} at iteration {i}"
                self.publish_work(
                    Task2WorkItem(processed_data=processed), input_task=task
                )

    class Task2Worker(TaskWorker):
        output_types: List[Type[Task]] = [Task3WorkItem]

        def consume_work(self, task: Task2WorkItem):
            self.print(f"Task2 consuming: {task.processed_data}")
            time.sleep(random.uniform(0.5, 2.5))

            if args.run_dashboard:
                # demonstrate the ability to request user input
                if random.random() < 0.15:
                    result, mime_type = self.request_user_input(
                        task=task,
                        instruction="Please provide a value",
                        accepted_mime_types=["text/html", "application/pdf"],
                    )
                    self.print(
                        f"User input: {len(result) if result else None} ({mime_type})"
                    )

            for i in range(11):
                final = f"Final: {task.processed_data} at iteration {i}!"
                self.publish_work(Task3WorkItem(final_result=final), input_task=task)

    class Task3Worker(TaskWorker):
        output_types: List[Type[Task]] = []

        def consume_work(self, task: Task3WorkItem):
            self.print(f"Task3 consuming: {task.final_result}")
            time.sleep(random.uniform(0.6, 1.2))
            self.print("Workflow complete!")

    # Create Graph
    graph = Graph(name="Simple Workflow")

    # Create tasks
    task1 = Task1Worker()
    task2 = Task2Worker()
    task3 = Task3Worker()

    # Add tasks to Graph
    graph.add_worker(task1).add_worker(task2).add_worker(task3)

    # Set dependencies
    graph.set_dependency(task1, task2).next(task3)

    # Prepare initial work item
    initial_work = [
        (task1, Task1WorkItem(data="Hello, Graph v1!")),
        (task1, Task1WorkItem(data="Hello, Graph v2!")),
    ]

    # Run the Graph
    graph.run(
        initial_work,
        run_dashboard=args.run_dashboard,
        dashboard_port=8080,
        display_terminal=not args.run_dashboard,
    )


if __name__ == "__main__":
    main()



# Copyright 2024 Niels Provos
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import logging
import os
import time
from contextlib import contextmanager
from datetime import datetime
from typing import Any, Dict, Optional
from xml.dom.minidom import parseString

import dicttoxml

# Suppress dicttoxml INFO logs at module initialization
dicttoxml_logger = logging.getLogger("dicttoxml")
dicttoxml_logger.setLevel(logging.WARNING)


def setup_logging(
    logs_dir: str = "logs",
    logs_prefix="general_log",
    logger_name: Optional[str] = None,
    level: Optional[int] = logging.INFO,
) -> logging.Logger:
    if not os.path.exists(logs_dir):
        os.makedirs(logs_dir)

    timestamp = datetime.now().strftime("%Y%m%d_%H%M%S")
    log_file = os.path.join(logs_dir, f"{logs_prefix}_{timestamp}.log")

    logger = logging.getLogger(name=logger_name)
    logger.handlers = []
    logger.setLevel(level)

    # Create file handler but don't create the file until the first emit
    file_handler = logging.FileHandler(log_file, mode="w", delay=True)
    file_handler.setLevel(level)

    # Create formatter and add it to the handler
    formatter = logging.Formatter(
        "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
    )
    file_handler.setFormatter(formatter)

    logger.addHandler(file_handler)

    return logger


@contextmanager
def measure_time():
    start_time = time.perf_counter()
    result = {}
    try:
        yield result
    finally:
        end_time = time.perf_counter()
        result["elapsed_time"] = (end_time - start_time) * 1000


def _is_valid_xml_char(char):
    """Checks if a character is valid in XML 1.0 (content)."""
    ord_val = ord(char)
    return (
        ord_val == 0x9  # Tab
        or ord_val == 0xA  # Line Feed
        or ord_val == 0xD  # Carriage Return
        or (0x20 <= ord_val <= 0xD7FF)
        or (0xE000 <= ord_val <= 0xFFFD)
        or (0x10000 <= ord_val <= 0x10FFFF)
    )


def _sanitize_key_for_xml(key: str) -> str:
    """Sanitizes dictionary keys to be valid XML tag names.
    Replaces invalid characters with '_'. Ensures key starts with a letter or underscore.
    """
    sanitized_key = ""
    for char in key:
        if char.isalnum() or char == "_":  # Allow alphanumeric and underscore
            sanitized_key += char
        else:
            sanitized_key += "_"  # Replace invalid key chars with underscore

    # Ensure key starts with a letter or underscore (XML tag requirement)
    if not (sanitized_key and (sanitized_key[0].isalpha() or sanitized_key[0] == "_")):
        sanitized_key = "_" + sanitized_key  # Prepend underscore if needed

    if not sanitized_key:
        return "_"  # Fallback if key becomes empty after sanitization

    return sanitized_key


def _sanitize_value_for_xml(value: Any) -> Any:
    """Sanitizes XML values while preserving type information."""
    if value is None or isinstance(value, (int, float, bool)):
        return value
    elif isinstance(value, (dict, list, tuple, set)):
        return _sanitize_for_xml(value)
    elif isinstance(value, bytes):
        try:
            value = value.decode("utf-8")
        except UnicodeDecodeError:
            value = str(value)
        return _sanitize_value_for_xml(value)
    elif isinstance(value, datetime):
        return value.isoformat()
    else:
        # Convert to string and sanitize only string values
        value = str(value)
        cleaned_string = ""
        for char in value:
            if _is_valid_xml_char(char):
                cleaned_string += char
            else:
                cleaned_string += "?"
        return cleaned_string


def _sanitize_for_xml(obj: Any) -> Any:
    """Recursively sanitize keys and values in a dictionary or list for XML."""
    if isinstance(obj, dict):
        return {
            _sanitize_key_for_xml(k): _sanitize_value_for_xml(v) for k, v in obj.items()
        }
    elif isinstance(obj, (list, tuple, set)):
        return [_sanitize_for_xml(item) for item in obj]
    else:
        return _sanitize_value_for_xml(obj)


def dict_dump_xml(dict: Dict[Any, Any], root: str = "root") -> str:
    """Formats the task as XML with sanitization and error handling."""
    # Sanitize the dictionary before conversion
    sanitized_dict = _sanitize_for_xml(dict)
    xml = dicttoxml.dicttoxml(
        sanitized_dict,
        custom_root=root,
        attr_type=False,
        item_func=lambda _: "item",  # Use "item" for all list items
    )
    # Decode bytes to string with utf-8 encoding
    xml_str = xml.decode("utf-8")

    xml_string = parseString(xml_str).toprettyxml(indent="  ")

    # Remove the XML declaration efficiently
    if xml_string.startswith("<?xml"):
        newline_index = xml_string.find("\n")
        if newline_index != -1:
            xml_string = xml_string[(newline_index + 1) :]
    return xml_string



# Copyright 2024 Niels Provos
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
dispatcher.py - Task Dispatching System

This module implements a sophisticated task dispatching system responsible for managing
the execution of tasks across multiple workers. It integrates with a separate provenance
tracking system to maintain a record of task lineage and dependencies.

Key Components:
- Dispatcher: The main class for handling task queues, worker assignment, and interaction
  with the provenance tracking system.
- ProvenanceTracker: External module responsible for maintaining the lineage of tasks.
- ProvenanceChain: Represents the lineage of a task as a tuple of (worker_name, task_id) pairs.
- TaskWorker: Base class for workers that process tasks.
- JoinedTaskWorker: A special type of worker that waits for multiple upstream tasks to complete.

Task Dispatching:
The Dispatcher manages tasks by assigning them to available workers, tracking their progress,
and ensuring tasks are completed in the correct order. It uses the following mechanisms:

1. Task Queues: Manages various queues for tasks, allowing for configurable parallelism limits
   on a per-worker type basis.

2. Task Lifecycle Management: Handles task state transitions from queued to active, and finally
   to completed or failed, capturing necessary statistics and logs.

3. User Input Handling: Provides facilities for tasks to request and receive user input asynchronously.

Integration with Provenance Tracking:
The Dispatcher relies on a separate ProvenanceTracker module to handle the intricacies of task
lineage and provenance. This enables:

1. Provenance Counting: Provenance count is externally managed to ensure accurate task tracking.

2. Notification Management: Ordered notifications for task dependencies are handled by the ProvenanceTracker,
   ensuring that all dependencies are satisfied before continuing to the next task.

Usage:
The Dispatcher class is the main interface for task management within PlanAI.
It coordinates with ProvenanceTracker to maintain the integrity of task dependencies while executing tasks
across various workers effectively.
"""


import logging
import random
import threading
import time
from collections import defaultdict, deque
from concurrent.futures import ThreadPoolExecutor
from queue import Empty, Queue
from threading import Event, RLock
from typing import TYPE_CHECKING, Any, Deque, Dict, List, Optional, Tuple, Type

from .provenance import ProvenanceChain
from .stats import WorkerStat
from .task import Task, TaskWorker
from .user_input import UserInputRequest
from .web_interface import is_quit_requested, run_web_interface, shutdown_web_interface

if TYPE_CHECKING:
    from .graph import Graph


class NotificationTask(Task):
    pass


def get_inheritance_chain(cls: Type[TaskWorker]) -> List[str]:
    """
    Returns the inheritance chain of a class as a list of strings.
    """
    if len(cls.__mro__) < 3:
        raise ValueError(
            "Class must be derived from TaskWorker and have at least 3 classes in its inheritance chain"
        )
    return [c.__name__ for c in cls.__mro__[:-2]]


class Dispatcher:
    def __init__(
        self, graph: "Graph", web_port: int = 5000, start_thread_pool: bool = True
    ):
        self._thread_pool = ThreadPoolExecutor() if start_thread_pool else None

        self._graphs = [graph]
        self._web_port = web_port
        self._web_thread = None

        self.task_lock = RLock()
        # We have a default Queue for all tasks
        self.work_queue = Queue()

        # And configurable queues for different worker classes
        # This allows us to reduce the maximum number of parallel tasks for specific worker classes, e.g. LLMs.
        self._per_worker_queue: Dict[str, Queue] = {}
        self._per_worker_task_count: Dict[str, int] = {}
        self._per_worker_max_parallel_tasks: Dict[str, int] = {}

        # we are using the work_available Event to signal the dispatcher that there might be work
        self.work_available = threading.Event()

        self.stop_event = Event()
        self.task_completion_event = Event()
        self._num_active_tasks = 0
        self._active_tasks: Dict[int, Tuple[TaskWorker, Task]] = {}
        self._aborted_prefixes: List[Tuple[ProvenanceChain, "Graph"]] = []
        self.completed_tasks: Deque[Tuple[TaskWorker, Task]] = deque(
            maxlen=100
        )  # Keep last 100 completed tasks
        self.failed_tasks: Deque[Tuple[TaskWorker, Task, str]] = deque(
            maxlen=100
        )  # Keep last 100 failed tasks
        self.worker_stats: Dict[str, WorkerStat] = defaultdict(
            WorkerStat
        )  # keeps track of execution times for each worker
        self.total_completed_tasks = 0
        self.total_failed_tasks = 0
        self.task_id_counter = 0

        # managing user requests
        self.user_input_requests = Queue()
        self.user_pending_requests: Dict[str, UserInputRequest] = {}

        # Add new log management
        self._log_queue = deque(maxlen=1000)  # Keep last 1000 log messages
        self._log_lock = RLock()

        self._shutdown_initiated = False

        self._dispatch_thread = None

        logging.info(
            "Dispatcher initialized with %d threads",
            self._thread_pool._max_workers if self._thread_pool else 0,
        )

    def register_graph(self, graph: "Graph"):
        if graph in self._graphs:
            raise ValueError("Graph already registered")
        self._graphs.append(graph)

    def deregister_graph(self, graph: "Graph"):
        if graph not in self._graphs:
            raise ValueError("Graph not registered")
        # we may need to consider removing the associated tasks from the queues
        # for now, we assume that this gets called after the main graph initiated a shutdown
        # on the dispatcher
        self._graphs.remove(graph)

    @property
    def active_tasks(self):
        with self.task_lock:
            return self._num_active_tasks

    def graph(self, index: int = 0) -> "Graph":
        if index >= len(self._graphs):
            raise ValueError(f"Graph index {index} out of range")
        return self._graphs[index]

    def _check_all_tasks_completed(self):
        with self.task_lock:
            if (
                self._num_active_tasks == 0
                and self.work_queue.empty()
                and all(q.empty() for q in self._per_worker_queue.values())
            ):
                return True
            return False

    def decrement_active_tasks(self, worker: TaskWorker) -> bool:
        """
        Decrements the count of active tasks by 1.

        Returns:
            bool: True if there are no more active tasks and the work queue is empty, False otherwise.
        """
        inherited_chain = get_inheritance_chain(worker.__class__)
        with self.task_lock:
            self.worker_stats[worker.name].decrement_active()
            for cls_name in inherited_chain:
                if cls_name in self._per_worker_task_count:
                    self._per_worker_task_count[cls_name] -= 1
                    if (
                        self._per_worker_task_count[cls_name]
                        < self._per_worker_max_parallel_tasks[cls_name]
                    ):
                        self.work_available.set()
            self._num_active_tasks -= 1
            if self._num_active_tasks < 0:
                raise ValueError("Number of active tasks is negative")
            if (
                self._num_active_tasks == 0
                and self.work_queue.empty()
                and all(q.empty() for q in self._per_worker_queue.values())
            ):
                logging.info("All tasks completed")
                return True
            return False

    def increment_active_tasks(self, worker: TaskWorker):
        inherited_chain = get_inheritance_chain(worker.__class__)
        with self.task_lock:
            self.worker_stats[worker.name].increment_active()
            for cls_name in inherited_chain:
                if cls_name in self._per_worker_task_count:
                    self._per_worker_task_count[cls_name] += 1
            self._num_active_tasks += 1

    def set_max_parallel_tasks(
        self, worker_class: Type[TaskWorker], max_parallel_tasks: int
    ):
        worker_class_name = worker_class.__name__
        with self.task_lock:
            if worker_class_name not in self._per_worker_queue:
                self._per_worker_queue[worker_class_name] = Queue()
                self._per_worker_task_count[worker_class_name] = 0
            self._per_worker_max_parallel_tasks[worker_class_name] = max_parallel_tasks

    def _dispatch_once(self) -> bool:
        # Add check for shutdown at the start
        if self._shutdown_initiated:
            return False

        # Check if there are user input requests
        self._dispatch_user_requests()

        queues = [(None, self.work_queue)] + list(self._per_worker_queue.items())
        random.shuffle(queues)
        for name, queue in queues:
            try:
                worker, task = queue.get_nowait()

                # Check if the task has been aborted
                if self._is_provenance_aborted(worker, task._provenance):
                    logging.info(
                        "Skipping task %s with %s due to aborted provenance chain",
                        task.name,
                        task._provenance,
                    )
                    worker._graph._provenance_tracker._remove_provenance(task, worker)
                    continue

                if name is not None:
                    # Check if the worker has reached its maximum parallel tasks
                    inheritance_chain = get_inheritance_chain(worker.__class__)
                    if name in inheritance_chain:
                        with self.task_lock:
                            if (
                                self._per_worker_task_count[name]
                                >= self._per_worker_max_parallel_tasks[name]
                            ):
                                queue.put((worker, task))
                                continue
                with self.task_lock:
                    self.worker_stats[worker.name].decrement_queued()

                # Use a named function instead of a lambda to avoid closure issues
                def task_completed_callback(future, worker=worker, task=task):
                    self._task_completed(worker, task, future)

                self.submit_work(
                    worker, [self._execute_task, worker, task], task_completed_callback
                )
                return True

            except Empty:
                pass  # this queue is empty, try the next one

        return False

    def submit_work(
        self,
        worker: TaskWorker,
        arguments: List[Any],
        task_completed_callback: callable,
    ):
        self.increment_active_tasks(worker)
        future = self._thread_pool.submit(*arguments)
        future.add_done_callback(task_completed_callback)

    def _dispatch_user_requests(self):
        """
        Handles the dispatching and processing of user input requests and their results.

        This function is responsible for the following:
        1. Checking for any pending user input requests and logging the instructions required
           for each. The requests are stored in a dictionary for tracking by their task ID.
        2. Checking for completed user input results, processing each result by matching it
           with its corresponding request, and then delivering the user's input back to the
           TaskWorker that requested it.

        The function operates in a non-blocking manner, retrieving requests and results from
        their respective queues only if they are available.
        """
        while not self.user_input_requests.empty():
            request: UserInputRequest = (
                self.user_input_requests.get_nowait()
            )  # Non-blocking retrieval
            logging.info(
                "User Input Required: Task ID %s - %s",
                request.task_id,
                request.instruction,
            )
            with self.task_lock:
                self.user_pending_requests[request.task_id] = request

    def dispatch(self):
        while True:
            # making sure that we can access active_tasks in a thread-safe way
            with self.task_lock:
                if (
                    self.stop_event.is_set()
                    and self.work_queue.empty()
                    and all(q.empty() for q in self._per_worker_queue.values())
                    and self._num_active_tasks == 0
                ) or self._shutdown_initiated:
                    logging.info("Stopping dispatcher")
                    break
            if self._dispatch_once():
                continue

            self.work_available.wait(timeout=0.1)
            self.work_available.clear()

    def _execute_task(self, worker: TaskWorker, task: Task):
        task_id = self._get_next_task_id()
        with self.task_lock:
            self._active_tasks[task_id] = (worker, task)

        # keep track of some basic timing information
        task._start_time = time.time()
        task._end_time = None

        try:
            # since we are storing a lot of references to the task, we need to make sure
            # that we are not storing the same task object in multiple places
            worker._pre_consume_work(task.model_copy())
        except Exception:
            raise  # Re-raise the caught exception
        finally:
            with self.task_lock:
                if task_id in self._active_tasks:
                    del self._active_tasks[task_id]

            task._end_time = time.time()

    def _get_next_task_id(self) -> int:
        with self.task_lock:
            self.task_id_counter += 1
            return self.task_id_counter

    def _task_to_dict(self, worker: TaskWorker, task: Task, error: str = "") -> Dict:
        data = {
            "id": self._get_task_id(task),
            "type": type(task).__name__,
            "worker": worker.name,
            "start_time": task._start_time,
            "end_time": task._end_time,
            "provenance": [f"{worker}_{id}" for worker, id in task._provenance],
            "input_provenance": [
                {input_task.name: input_task.model_dump()}
                for input_task in task._input_provenance
            ],
        }
        if error:
            data["error"] = error
        return data

    def get_graphs(self) -> List["Dict"]:
        return [
            {"index": index, "name": graph.name}
            for index, graph in enumerate(self._graphs)
        ]

    def get_traces(self, index: int = 0) -> Dict:
        if index >= len(self._graphs):
            return {}
        return self._graphs[index]._provenance_tracker.get_traces()

    def get_queued_tasks(self) -> List[Dict]:
        work_items = []
        with self.task_lock:
            work_items.extend(self.work_queue.queue)
            for queue in self._per_worker_queue.values():
                work_items.extend(queue.queue)
            return [self._task_to_dict(worker, task) for worker, task in work_items]

    def get_active_tasks(self) -> List[Dict]:
        with self.task_lock:
            return [
                self._task_to_dict(worker, task)
                for task_id, (worker, task) in self._active_tasks.items()
            ]

    def get_completed_tasks(self) -> List[Dict]:
        with self.task_lock:
            return [
                self._task_to_dict(worker, task)
                for worker, task in self.completed_tasks
            ]

    def get_failed_tasks(self) -> List[Dict]:
        with self.task_lock:
            return [
                self._task_to_dict(worker, task, error)
                for worker, task, error in self.failed_tasks
            ]

    def get_execution_statistics(self):
        stats = {
            worker: stat.get_statistics() for worker, stat in self.worker_stats.items()
        }
        return stats

    def get_user_input_requests(self) -> List[Dict]:
        with self.task_lock:
            return [
                {
                    "task_id": request.task_id,
                    "instruction": request.instruction,
                    "accepted_mime_types": request.accepted_mime_types,
                }
                for request in self.user_pending_requests.values()
            ]

    def set_user_input_result(
        self, task_id: str, result: Any, mime_type: Optional[str] = None
    ):
        logging.info(
            "User Input Received: Task ID %s - Result: %s (MIME Type: %s)",
            task_id,
            result[:30] if result else "<None>",
            mime_type,
        )

        # Locate the request for task_id and inform the TaskWorker
        with self.task_lock:
            if task_id in self.user_pending_requests:
                request: UserInputRequest = self.user_pending_requests.pop(task_id)
                # Provide the result to the requesting TaskWorker's queue
                request.respond(
                    result,
                    mime_type,
                )

    def _get_task_id(self, task: Task) -> str:
        # Use the last entry in the _provenance list as the task ID
        if task._provenance:
            return f"{task._provenance[-1][0]}_{task._provenance[-1][1]}"
        else:
            # Fallback in case _provenance is empty
            return f"unknown_{id(task)}"

    def create_work_hold(self):
        """Artificially increment the number of active tasks to prevent completion."""
        with self.task_lock:
            self._num_active_tasks += 1

    def release_work_hold(self):
        """Release the artificial hold we created and check whether work remains."""
        with self.task_lock:
            self._num_active_tasks -= 1
            if self._num_active_tasks == 0 and self._check_all_tasks_completed():
                self.task_completion_event.set()

    def _task_completed(self, worker: TaskWorker, task: Optional[Task], future):
        success: bool = False
        error_message: str = ""
        try:
            # This will raise an exception if the task failed
            _ = future.result()

            # Handle successful task completion
            if task:
                logging.info(f"Task {task.name} completed successfully")
            else:
                logging.info(
                    f"Notification for worker {worker.name} completed successfully"
                )
            success = True

            # collect execution stats
            if task and task._start_time and task._end_time:
                execution_time = task._end_time - task._start_time
                with self.task_lock:
                    self.worker_stats[worker.name].add_completion_time(execution_time)

        except Exception as e:
            # Handle task failure
            error_message = str(e)
            if task:
                logging.exception(
                    "Worker %s on Task %s failed with exception: %s",
                    worker.name,
                    task.name,
                    str(e),
                )
            else:
                logging.exception(
                    f"Notification for worker {worker.name} failed with exception: {str(e)}"
                )

            # Anything else that needs to be done when a task fails?

        finally:
            # This code will run whether the task succeeded or failed
            if not success:
                if task is not None:
                    # Determine whether we should retry the task
                    if worker.num_retries > 0:
                        if task.retry_count < worker.num_retries:
                            task.increment_retry_count()
                            self.decrement_active_tasks(worker)
                            self._add_to_queue(worker, task)
                            logging.info(
                                f"Retrying task {task.name} for the {task.retry_count} time"
                            )
                            return

                with self.task_lock:
                    self.failed_tasks.appendleft(
                        (worker, task if task else NotificationTask(), error_message)
                    )
                    self.worker_stats[worker.name].increment_failed()
                    self.total_failed_tasks += 1

                if task:
                    logging.error(
                        f"Task {task.name} failed after {task.retry_count} retries"
                    )
            else:
                with self.task_lock:
                    self.completed_tasks.appendleft(
                        (worker, task if task else NotificationTask())
                    )
                    self.total_completed_tasks += 1
                    self.worker_stats[worker.name].increment_completed()

            if task:
                worker._graph._provenance_tracker._remove_provenance(task, worker)

            if self.decrement_active_tasks(worker):
                self.task_completion_event.set()

    def abort_work(self, graph: "Graph", ProvenanceChain: List[Tuple[str, str]]):
        with self.task_lock:
            self._aborted_prefixes.append((tuple(ProvenanceChain), graph))

    def _is_provenance_aborted(
        self, worker: TaskWorker, provenance: ProvenanceChain
    ) -> bool:
        """
        Checks if any prefix in self._aborted_prefixes matches the given provenance chain.

        Args:
            provenance: The provenance chain to check

        Returns:
            bool: True if the provenance chain has an aborted prefix
        """
        with self.task_lock:
            for aborted_prefix, graph in self._aborted_prefixes:
                if worker._graph == graph and provenance[: len(aborted_prefix)] == list(
                    aborted_prefix
                ):
                    return True
        return False

    def add_work(self, worker: TaskWorker, task: Task):
        # Skip adding the task if its provenance chain has been aborted
        if self._is_provenance_aborted(worker, task._provenance):
            logging.info("Skipping task %s due to aborted provenance chain", task.name)
            return

        task_copy = task.model_copy()
        worker._graph._provenance_tracker._add_provenance(task_copy)

        self._add_to_queue(worker, task_copy)

    def add_multiple_work(self, work_items: List[Tuple[TaskWorker, Task]]):
        for worker, task in work_items:
            self.add_work(worker, task)

    def _add_to_queue(self, worker: TaskWorker, task: Task):
        if self._shutdown_initiated:
            return

        inheritance_chain = get_inheritance_chain(worker.__class__)
        per_task_queue = self.work_queue
        with self.task_lock:
            self.worker_stats[worker.name].increment_queued()
            for cls_name in inheritance_chain:
                if cls_name in self._per_worker_queue:
                    per_task_queue = self._per_worker_queue[cls_name]
                    break
        per_task_queue.put((worker, task))
        self.work_available.set()

    def wait_for_completion(self, timeout: float = None, wait_for_quit=False):
        """
        Wait for all tasks to complete or timeout.

        Args:
            timeout (float): Maximum time to wait in seconds
            wait_for_quit (bool): Whether to wait for quit request
        """
        completed = self.task_completion_event.wait(timeout=timeout)
        if not completed:
            logging.warning("Dispatcher wait_for_completion timed out")
            wait_for_quit = False

        if wait_for_quit:
            while not is_quit_requested():
                # Sleep for a short time to avoid busy waiting
                time.sleep(0.1)

        if self._thread_pool:
            logging.info("Shutting down thread pool")
            self._thread_pool.shutdown(wait=True, cancel_futures=True)
            logging.info("Thread pool shut down")
            self._thread_pool = None

        if self._web_thread:
            logging.info("Stopping web thread")
            try:
                shutdown_web_interface()
            except Exception as e:
                logging.warning(f"Error shutting down web interface: {e}")
            self._web_thread.join(timeout=0.5)
            self._web_thread = None

    def has_dashboard(self):
        """
        Check if the dashboard web thread is running.

        Returns:
            bool: True if the dashboard web thread is running, False otherwise.
        """
        return self._web_thread is not None

    def start_web_interface(self):
        self._web_thread = threading.Thread(
            target=run_web_interface, args=(self, self._web_port)
        )
        self._web_thread.daemon = (
            True  # This ensures the web thread will exit when the main thread exits
        )
        self._web_thread.start()

    def get_user_input_requests_queue(self) -> Queue:
        """Get the user input requests queue."""
        return self.user_input_requests

    def add_log(self, message: str) -> None:
        """Add a log message to the queue."""
        with self._log_lock:
            self._log_queue.append({"timestamp": time.time(), "message": message})

    def get_logs(self) -> List[Dict]:
        """Get all logs and clear the queue."""
        with self._log_lock:
            logs = list(self._log_queue)
            self._log_queue.clear()
            return logs

    def initiate_shutdown(self):
        """Initiates a graceful shutdown of the dispatcher."""
        self._shutdown_initiated = True
        self.task_completion_event.set()
        logging.info("Dispatcher shutdown initiated")

    def start(self):
        """Start the dispatcher thread if not already running."""
        if self._thread_pool is None:
            self._thread_pool = ThreadPoolExecutor()
        if self._dispatch_thread is None:
            self._dispatch_thread = threading.Thread(target=self.dispatch)
            self._dispatch_thread.daemon = True
            self._dispatch_thread.start()
            logging.info("Dispatcher thread started")
        else:
            logging.warning("Dispatcher thread already running")

    def stop(self, timeout: float = None):
        """
        Stop the dispatcher and wait for thread completion.

        Args:
            timeout (float, optional): Maximum time to wait for thread completion in seconds
        """
        self.stop_event.set()
        if self._dispatch_thread:
            self._dispatch_thread.join(timeout=timeout)
            if self._dispatch_thread.is_alive():
                logging.warning("Dispatcher thread did not stop within timeout")
            else:
                logging.info("Dispatcher thread stopped")
            self._dispatch_thread = None
